@@ -10,6 +10,7 @@ import parallel as par
 
 import atom3.database as db
 import atom3.sequence as sequ
+from atom3.structure import get_chain_to_valid_residues
 
 
 def add_conservation_parser(subparsers, pp):
@@ -34,6 +35,53 @@ def add_conservation_parser(subparsers, pp):
                          ' number processors available on current machine)')
 
 
+def gen_protrusion_index(pdb_filename, output_filename):
+    """Generate protrusion index from structure."""
+    pdb_name = db.get_pdb_name(pdb_filename)
+    out_dir = os.path.dirname(output_filename)
+    work_dir = os.path.join(out_dir, 'work')
+    if not os.path.exists(work_dir):
+        os.makedirs(work_dir)
+    structure = pd.read_pickle(pdb_filename)
+    chains = [chain_res_tuple[0][1] for chain_res_tuple in get_chain_to_valid_residues(structure)]
+    pdb_names = [pdb_name for _ in range(len(chains))]
+
+    protrusion_indices = []
+    for chain, pdb_name in zip(chains, pdb_names):
+        basename = os.path.splitext(pdb_name)[0]
+        protrusion_index_filename = "{}.psaia".format(basename)
+        output = os.path.join(work_dir, basename)
+        if not os.path.exists(protrusion_index_filename):
+            logging.info("PSAIA'ing {:}".format(pdb_name))
+            _psaia(pdb_name, protrusion_index_filename, output)
+
+        if not os.path.exists(protrusion_index_filename):
+            logging.warning("No hits for {:}".format(pdb_name))
+            # Create empty file.
+            open(protrusion_index_filename, 'w').close()
+
+        if os.stat(protrusion_index_filename).st_size != 0:
+            protrusion_indices = pd.read_csv(
+                protrusion_index_filename, skiprows=2, skipfooter=6, delim_whitespace=True,
+                engine='python', usecols=range(20), index_col=[0, 1])
+            protrusion_indices = protrusion_indices.reset_index()
+            del protrusion_indices['level_0']
+            protrusion_indices.rename(columns={'level_1': 'orig'}, inplace=True)
+            del protrusion_indices['orig']
+
+        else:
+            logging.warning("No protrusion index found for {:} (model {:}, chain {:})"
+                            .format(pdb_name, chain[-2], chain[-1]))
+            protrusion_indices = None
+
+        protrusion_indices['pdb_name'] = db.get_pdb_name(pdb_filename)
+        protrusion_indices['model'] = chain[0]
+        protrusion_indices['chain'] = chain[1]
+        protrusion_indices.append(protrusion_indices)
+    protrusion_indices = pd.concat(protrusion_indices)
+    return protrusion_indices
+
+
 def gen_pssm(pdb_filename, blastdb, output_filename):
     """Generate PSSM and PSFM from sequence."""
     pdb_name = db.get_pdb_name(pdb_filename)
@@ -116,6 +164,28 @@ def gen_pssm(pdb_filename, blastdb, output_filename):
     return pssms
 
 
+def map_protrusion_indices(pdb_filename, output_filename):
+    pdb_name = db.get_pdb_name(pdb_filename)
+    start_time = timeit.default_timer()
+    start_time_blasting = timeit.default_timer()
+    pis = gen_protrusion_index(pdb_filename, output_filename)
+    num_chains = len(pis.groupby(['pdb_name', 'model', 'chain']))
+    elapsed_blasting = timeit.default_timer() - start_time_blasting
+
+    parsed = pd.read_pickle(pdb_filename)
+    parsed = parsed.merge(
+        pis, on=['model', 'pdb_name', 'chain', 'residue'])
+
+    start_time_writing = timeit.default_timer()
+    parsed.to_pickle(output_filename)
+    elapsed_writing = timeit.default_timer() - start_time_writing
+
+    elapsed = timeit.default_timer() - start_time
+    logging.info(('For {:d} pssms generated from {} spent {:05.2f} blasting,'
+                  ' {:05.2f} writing, and {:05.2f} overall.').format(num_chains, pdb_name, elapsed_blasting,
+                                                                     elapsed_writing, elapsed))
+
+
 def map_pssms(pdb_filename, blastdb, output_filename):
     pdb_name = db.get_pdb_name(pdb_filename)
     start_time = timeit.default_timer()
@@ -138,6 +208,20 @@ def map_pssms(pdb_filename, blastdb, output_filename):
                                                                      elapsed_writing, elapsed))
 
 
+def _psaia(query, output_protrusion_index, output):
+    """Run PSAIA on specified input."""
+    psaia_command = "psa {:} {:}"  # PSA is the CLI for PSAIA (the GUI)
+    log_out = "{}.out".format(output)
+    log_err = "{}.err".format(output)
+    with open(log_out, 'a') as f_out:
+        with open(log_err, 'a') as f_err:
+            command = psaia_command.format(query, output + '.psaia', output_protrusion_index)
+            f_out.write('=================== CALL ===================\n')
+            f_out.write(command + '\n')
+            subprocess.check_call(command, shell=True, stderr=f_err, stdout=f_out)
+            f_out.write('================= END CALL =================\n')
+
+
 def _blast(query, output_pssm, output, blastdb):
     """Run PSIBlast on specified input."""
     psiblast_command = "psiblast -db {:} -query {:} -out_ascii_pssm {:} " + \
@@ -146,12 +230,10 @@ def _blast(query, output_pssm, output, blastdb):
     log_err = "{}.err".format(output)
     with open(log_out, 'a') as f_out:
         with open(log_err, 'a') as f_err:
-            command = psiblast_command.format(
-                blastdb, query, output_pssm, output)
+            command = psiblast_command.format(blastdb, query, output_pssm, output)
             f_out.write('=================== CALL ===================\n')
             f_out.write(command + '\n')
-            subprocess.check_call(
-                command, shell=True, stderr=f_err, stdout=f_out)
+            subprocess.check_call(command, shell=True, stderr=f_err, stdout=f_out)
             f_out.write('================= END CALL =================\n')
 
 
@@ -165,8 +247,7 @@ def _to_clustal(psiblast_in, clustal_out):
             command = mview_command.format(psiblast_in, clustal_out)
             f_out.write('=================== CALL ===================\n')
             f_out.write(command + '\n')
-            subprocess.check_call(
-                command, shell=True, stderr=f_err, stdout=f_out)
+            subprocess.check_call(command, shell=True, stderr=f_err, stdout=f_out)
             f_out.write('================= END CALL =================\n')
 
 
@@ -180,11 +261,35 @@ def _al2co(clustal_in, al2co_out):
             command = al2co_command.format(clustal_in, al2co_out)
             f_out.write('=================== CALL ===================\n')
             f_out.write(command + '\n')
-            subprocess.check_call(
-                command, shell=True, stderr=f_err, stdout=f_out)
+            subprocess.check_call(command, shell=True, stderr=f_err, stdout=f_out)
             f_out.write('================= END CALL =================\n')
 
 
+def map_all_protrusion_indices(pdb_dataset, output_dir, num_cpus):
+    ext = '.pkl'
+    requested_filenames = \
+        db.get_structures_filenames(pdb_dataset, extension=ext)
+    requested_keys = [db.get_pdb_name(x) for x in requested_filenames]
+    produced_filenames = db.get_structures_filenames(
+        output_dir, extension='.pkl')
+    produced_keys = [db.get_pdb_name(x) for x in produced_filenames]
+    work_keys = [key for key in requested_keys if key not in produced_keys]
+    work_filenames = [x[0] for x in db.get_all_filenames(work_keys, pdb_dataset, extension=ext,
+                                                         keyer=lambda x: db.get_pdb_name(x))]
+
+    output_filenames = []
+    for pdb_filename in work_filenames:
+        sub_dir = output_dir + '/' + db.get_pdb_code(pdb_filename)[1:3]
+        if not os.path.exists(sub_dir):
+            os.makedirs(sub_dir)
+        output_filenames.append(sub_dir + '/' + db.get_pdb_name(pdb_filename) + ".pkl")
+
+    logging.info("{:} requested keys, {:} produced keys, {:} work keys".format(len(requested_keys),
+                                                                               len(produced_keys), len(work_keys)))
+    inputs = [(key, output) for key, output in zip(work_filenames, output_filenames)]
+    par.submit_jobs(map_protrusion_indices, inputs, num_cpus)
+
+
 def map_all_pssms(pdb_dataset, blastdb, output_dir, num_cpus):
     ext = '.pkl'
     requested_filenames = \
